/*
 *  -- MAGMA (version 1.1) --
 *     Univ. of Tennessee, Knoxville
 *     Univ. of California, Berkeley
 *     Univ. of Colorado, Denver
 *     April 2012
 *
 * @generated d Fri Nov 22 14:21:31 2013
 *
 **/
// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "magma_types.h"
#include "testings.h"


/* ////////////////////////////////////////////////////////////////////////////
   -- Testing dtrmm
*/
int main( int argc, char** argv)
{
    /* Initialize */
    magma_queue_t  queue;
    magma_device_t device;
    int num = 0;
    magma_err_t err;
    magma_init();
    err = magma_get_devices( &device, 1, &num );
    if ( err != MAGMA_SUCCESS or num < 1 ) {
      fprintf( stderr, "magma_get_devices failed: %d\n", err );
      exit(-1);
    }
    err = magma_queue_create( device, &queue );
    if ( err != MAGMA_SUCCESS ) {
      fprintf( stderr, "magma_queue_create failed: %d\n", err );
      exit(-1);
    }

    real_Double_t   gflops, mic_perf, mic_time, cpu_perf, cpu_time;
    double          mic_error, Cnorm, work[1];
    magma_int_t M, N;
    magma_int_t Ak;
    magma_int_t sizeA, sizeB;
    magma_int_t lda, ldb, ldda, lddb;
    magma_int_t ione     = 1;
    magma_int_t ISEED[4] = {0,0,0,1};
    
    double *h_A, *h_B, *h_Bmic;
    double *d_A, *d_B;
    double c_neg_one = MAGMA_D_NEG_ONE;
    double alpha = MAGMA_D_MAKE(  0.29, -0.86 );
    
    magma_opts opts;
    parse_opts( argc, argv, &opts );
    
    printf("    M     N   MIC Gflop/s (ms)   CPU Gflop/s (ms)  MIC error\n");
    printf("==================================================================\n");
    for( int i = 0; i < opts.ntest; ++i ) {
        for( int iter = 0; iter < opts.niter; ++iter ) {
            M = opts.msize[i];
            N = opts.nsize[i];
            gflops = FLOPS_DTRMM(opts.side, M, N) / 1e9;

            if ( opts.side == MagmaLeft ) {
                lda = M;
                Ak = M;
            } else {
                lda = N;
                Ak = N;
            }
            
            ldb = M;
            
            ldda = ((lda+31)/32)*32;
            lddb = ((ldb+31)/32)*32;
            
            sizeA = lda*Ak;
            sizeB = ldb*N;
            
            TESTING_MALLOC_HOST( h_A,  double, lda*Ak );
            TESTING_MALLOC_HOST( h_B,  double, ldb*N  );
            TESTING_MALLOC_HOST( h_Bmic, double, ldb*N  );
            
            TESTING_MALLOC_DEV( d_A, double, ldda*Ak );
            TESTING_MALLOC_DEV( d_B, double, lddb*N  );
            
            /* Initialize the matrices */
            lapackf77_dlarnv( &ione, ISEED, &sizeA, h_A );
            lapackf77_dlarnv( &ione, ISEED, &sizeB, h_B );
            
            /* =====================================================================
               Performs operation using CUDA-BLAS
               =================================================================== */
            magma_dsetmatrix( Ak, Ak, h_A, 0, lda, d_A, 0, ldda, queue );
            magma_dsetmatrix( M, N, h_B, 0, ldb, d_B, 0, lddb, queue );
            
            mic_time = magma_sync_wtime( NULL );
            magma_dtrmm( opts.side, opts.uplo, opts.transA, opts.diag,
                         M, N, 
                         alpha, d_A, 0, ldda,
                         d_B, 0, lddb, queue );
            mic_time = magma_sync_wtime( NULL ) - mic_time;
            mic_perf = gflops / mic_time;
            
            
	    /* =====================================================================
	       Performs operation using CPU BLAS
	       =================================================================== */
	    if ( opts.check ) {
		    magma_dgetmatrix( M, N, d_B, 0, lddb, h_Bmic, 0, ldb, queue );
		    cpu_time = magma_wtime();
		    blasf77_dtrmm( lapack_const(opts.side), lapack_const(opts.uplo), lapack_const(opts.transA), lapack_const(opts.diag), 
				    &M, &N,
				    &alpha, h_A, &lda,
				    h_B, &ldb );
		    cpu_time = magma_wtime() - cpu_time;
		    cpu_perf = gflops / cpu_time;

		    /* =====================================================================
		       Check the result
		       =================================================================== */
		    // compute relative error for both magma & mic, relative to lapack,
		    // |C_magma - C_lapack| / |C_lapack|
		    Cnorm = lapackf77_dlange( "M", &M, &N, h_B, &ldb, work );

		    blasf77_daxpy( &sizeB, &c_neg_one, h_B, &ione, h_Bmic, &ione );
		    mic_error = lapackf77_dlange( "M", &M, &N, h_Bmic, &ldb, work ) / Cnorm;

		    printf("%5d %5d   %7.2f (%7.2f)   %7.2f (%7.2f)    %8.2e\n",
				    (int) M, (int) N,
				    mic_perf, 1000.*mic_time,
				    cpu_perf,    1000.*cpu_time,
				    mic_error );
	    }
	    else {
		    printf("%5d %5d   %7.2f (%7.2f)    ---   (  ---  )    ---\n",
				    (int) M, (int) N,
				    mic_perf, 1000.*mic_time);
	    }

	    TESTING_FREE_HOST( h_A  );
	    TESTING_FREE_HOST( h_B  );
	    TESTING_FREE_HOST( h_Bmic );

	    TESTING_FREE_DEV( d_A );
	    TESTING_FREE_DEV( d_B );
	}
	if ( opts.niter > 1 ) {
		printf( "\n" );
	}
    }

    magma_queue_destroy( queue );    
    magma_finalize();

    return 0;
}
